from typing import Type

import pandas as pd
import plotly.graph_objects as go
from django.db.models.expressions import F
from django.db.models.query import QuerySet
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from rest_framework import filters
from wbcore import viewsets
from wbcore.contrib.pandas import fields as pf
from wbcore.contrib.pandas.views import PandasAPIViewSet
from wbcore.filters import DjangoFilterBackend
from wbcore.serializers.serializers import ModelSerializer
from wbcore.utils.strings import format_number

from wbhuman_resources.filters import (
    KPIEvaluationFilterSet,
    KPIEvaluationPandasFilter,
    KPIFilterSet,
)
from wbhuman_resources.models import KPI, Evaluation
from wbhuman_resources.serializers import (
    EvaluationModelSerializer,
    EvaluationRepresentationSerializer,
    KPIModelSerializer,
    KPIRepresentationSerializer,
)
from wbhuman_resources.viewsets.buttons import KPIButtonConfig
from wbhuman_resources.viewsets.display import (
    KPIDisplayConfig,
    KPIEvaluationDisplayConfig,
    KPIEvaluationPandasDisplayConfig,
)
from wbhuman_resources.viewsets.endpoints import (
    EvaluationGraphEndpointConfig,
    KPIEndpointConfig,
    KPIEvaluationEndpointConfig,
    KPIEvaluationPandasEndpointConfig,
)
from wbhuman_resources.viewsets.titles import (
    EvaluationGraphTitleConfig,
    KPIEvaluationPandasTitleConfig,
)


class KPIRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = KPI.objects.all()
    serializer_class = KPIRepresentationSerializer


class KPIModelViewSet(viewsets.ModelViewSet):
    queryset = KPI.objects.all()
    serializer_class = KPIModelSerializer
    display_config_class = KPIDisplayConfig
    endpoint_config_class = KPIEndpointConfig
    button_config_class = KPIButtonConfig
    filter_backends = (
        filters.SearchFilter,
        filters.OrderingFilter,
        DjangoFilterBackend,
    )
    filterset_class = KPIFilterSet
    ordering = ["last_update", "period"]
    ordering_fields = ["name", "evaluated_persons", "parameters", "period"]
    search_fields = ["name", "evaluated_persons__computed_str"]

    def get_serializer_class(self) -> Type[ModelSerializer]:
        if pk := self.kwargs.get("pk", None):
            kpi = get_object_or_404(KPI, pk=pk)
            handler = kpi.get_handler()
            return handler.get_serializer()
        return super().get_serializer_class()

    def get_queryset(self) -> QuerySet[KPI]:
        queryset = super().get_queryset()
        if pk := self.kwargs.get("pk", None):
            kpi = get_object_or_404(KPI, pk=pk)
            handler = kpi.get_handler()
            queryset = handler.annotate_parameters(queryset)
        queryset = queryset.annotate(
            parameters=F("additional_data__list_data"),
        )
        if KPI.is_administrator(self.request.user):
            return queryset
        return queryset.filter(evaluated_persons=self.request.user.profile)


class EvaluationRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = Evaluation.objects.all()
    serializer_class = EvaluationRepresentationSerializer


class KPIEvaluationModelViewSet(viewsets.ModelViewSet):
    serializer_class = EvaluationModelSerializer
    display_config_class = KPIEvaluationDisplayConfig
    endpoint_config_class = KPIEvaluationEndpointConfig
    queryset = Evaluation.objects.all()
    filter_backends = (
        filters.SearchFilter,
        filters.OrderingFilter,
        DjangoFilterBackend,
    )
    filterset_class = KPIEvaluationFilterSet
    ordering = ["-evaluated_period"]
    ordering_fields = ["person", "evaluated_score", "evaluation_date", "evaluated_period"]

    def get_queryset(self):
        queryset = super().get_queryset().filter(kpi__id=self.kwargs["kpi_id"]).annotate(goal=F("kpi__goal"))
        if KPI.is_administrator(self.request.user):
            return queryset
        return queryset.filter(person=self.request.user.profile)

    def get_aggregates(self, queryset, paginated_queryset):
        if queryset.exists():
            kpi = get_object_or_404(KPI, pk=self.kwargs["kpi_id"])
            person_ids = kpi.evaluated_persons.all().values_list("id", flat=True)
            evaluation = queryset.filter(person__id__in=person_ids).order_by("evaluation_date").last()
            evaluated_score = kpi.get_handler().evaluate(kpi, evaluation_date=evaluation.evaluation_date)
            return {
                "evaluated_score": {"Σ": format_number(evaluated_score)},
            }
        return {}


class EvaluationGraphChartViewset(viewsets.ChartViewSet):
    filter_backends = (DjangoFilterBackend,)
    queryset = Evaluation.objects.all()
    endpoint_config_class = EvaluationGraphEndpointConfig
    title_config_class = EvaluationGraphTitleConfig
    # filterset_class = EvaluationGraphFilter

    def get_queryset(self):
        queryset = super().get_queryset().filter(kpi__id=self.kwargs["kpi_id"]).annotate(goal=F("kpi__goal"))
        if KPI.is_administrator(self.request.user):
            return queryset
        return queryset.filter(person=self.request.user.profile)

    def get_dataframe_period(self, kpi_id):
        kpi = KPI.objects.get(id=kpi_id)
        list_date = list(
            pd.date_range(
                start=kpi.period.lower,
                end=kpi.period.upper,
                freq=KPI.Interval.get_frequence_correspondance(kpi.evaluated_intervals),
            )
        )
        list_date = [_date.date() for _date in list_date]

        df_date = pd.DataFrame(list_date, columns=["evaluation_date"])
        df_goal = df_date.copy()
        df_goal["goal"] = kpi.goal
        df_date = df_date.set_index("evaluation_date")
        return df_date, df_goal["goal"]

    def get_plotly(self, queryset):
        fig = go.Figure()
        if queryset.exists():
            kpi = KPI.objects.get(id=self.kwargs["kpi_id"])
            activity_type = kpi.additional_data.get("list_data").pop(0) if kpi.additional_data.get("list_data") else ""
            parameters = kpi.additional_data.get("list_data")

            df = pd.DataFrame(
                queryset.values(
                    "evaluation_date", "person__first_name", "person__last_name", "evaluated_score", "goal"
                )
            )
            df["Name"] = df["person__first_name"] + " " + df["person__last_name"]
            df["Name"] = df["Name"].fillna("Group")
            df = df.set_index(["Name", "evaluation_date"]).evaluated_score.unstack("Name")
            df_date, goal = self.get_dataframe_period(self.kwargs["kpi_id"])

            df = df.merge(df_date, on=["evaluation_date"], how="outer")
            df = df.sort_values(by=["evaluation_date"]).fillna("")
            pd.options.plotting.backend = "plotly"
            fig = df.plot.line(labels=dict(value="#KPI", variable="Name"), markers=True)
            fig.update_traces(hovertemplate=None)
            fig.update_layout(
                title=gettext("<b>Parameters</b>: {}<br>{} ").format(parameters, activity_type),
                hovermode="x",
                xaxis_rangeslider_visible=True,
                autosize=True,
            )
            fig.update_xaxes(title="")
            fig.add_trace(  # goal line
                go.Scatter(
                    x=df.index,
                    y=goal,
                    name=gettext("Goal"),
                    mode="lines",
                    line=dict(color="Green", width=4, dash="dot"),
                )
            )
            fig.add_trace(  # diagonal line
                go.Scatter(
                    x=df.index,
                    y=(goal / len(df)).cumsum().round(2),
                    mode="lines",
                    name=gettext("Progress expected to reach goal"),
                    line=dict(color="springgreen", width=3, dash="dot"),
                )
            )
        return fig


class KPIEvaluationPandasViewSet(PandasAPIViewSet):
    IDENTIFIER = "wbcommission:kpievaluationpandas"
    queryset = Evaluation.objects.all()

    filter_backends = (
        DjangoFilterBackend,
        filters.SearchFilter,
        filters.OrderingFilter,
    )

    search_fields = ["kpi__name", "person__computed_str"]
    ordering_fields = [
        "evaluated_period",
        "evaluation_date",
    ]

    display_config_class = KPIEvaluationPandasDisplayConfig
    title_config_class = KPIEvaluationPandasTitleConfig
    endpoint_config_class = KPIEvaluationPandasEndpointConfig
    filterset_class = KPIEvaluationPandasFilter

    def get_queryset(self):
        queryset = super().get_queryset().annotate(goal=F("kpi__goal"), kpi_name=F("kpi__name"))
        if KPI.is_administrator(self.request.user):
            return queryset
        return queryset.filter(kpi__evaluated_persons=self.request.user.profile)

    def get_pandas_fields(self, request):
        fields = [
            pf.PKField("id", label=_("ID")),
            pf.CharField(key="person", label=_("Person")),
            pf.CharField(key="kpi", label=_("KPI")),
            pf.CharField(key="kpi_name", label=_("KPI")),
            pf.CharField(key="period", label=_("Period")),
            pf.DateField(key="evaluation_date", label=_("Evaluation Date")),
            pf.IntegerField(key="goal", label=_("Goal")),
            pf.IntegerField(key="evaluated_score", label=_("Evaluated Score")),
            pf.CharField(key="progression", label=_("Progression")),
        ]
        return pf.PandasFields(fields=tuple(fields))

    def get_dataframe(self, request, queryset, **kwargs):
        df = pd.DataFrame()
        if queryset.exists():
            df = pd.DataFrame(
                queryset.order_by("evaluation_date").values(
                    "kpi",
                    "kpi_name",
                    "kpi__period",
                    "evaluation_date",
                    "goal",
                    "person__computed_str",
                    "evaluated_score",
                )
            )
            df["kpi__period"] = df["kpi__period"].apply(
                lambda x: f"{x.lower.strftime('%d.%m.%Y')} - {x.upper.strftime('%d.%m.%Y')}"
            )

            df["person__computed_str"] = df["person__computed_str"].fillna("Group").astype(str)
            df = df.rename(columns={"kpi__period": "period", "person__computed_str": "person"})
            df = df.pivot_table(
                index=["person", "kpi"],
                values=["kpi_name", "evaluation_date", "period", "goal", "evaluated_score"],
                aggfunc="last",
            )
            if "evaluated_score" in df.columns:
                df["progression"] = ((df["evaluated_score"] / df["goal"]) * 100).round(2).astype(str) + "%"
            else:
                df["progression"] = "0%"

            df = df.reset_index()
            df["id"] = df.index
        return df



from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property

from wbhuman_resources.models.employee import EmployeeHumanResource


class EmployeeViewMixin:
    @cached_property
    def employee(self):
        return get_object_or_404(EmployeeHumanResource, profile=self.request.user.profile)

    @cached_property
    def is_administrator(self):
        return EmployeeHumanResource.is_administrator(self.request.user)



import random
from datetime import timedelta

import factory
import pytz
from django.utils import timezone
from faker import Faker
from psycopg.types.range import DateRange
from wbcore.contrib.authentication.factories import InternalUserFactory
from wbcore.contrib.directory.factories.entries import PersonFactory

from wbhuman_resources.models import (
    KPI,
    Evaluation,
    Review,
    ReviewAnswer,
    ReviewGroup,
    ReviewQuestion,
    ReviewQuestionCategory,
)

fake = Faker()


class ReviewGroupFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = ReviewGroup
        skip_postgeneration_save = True

    name = factory.Faker("text", max_nb_chars=64)

    @factory.post_generation
    def employees(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for employee in extracted:
                self.employees.add(employee)


class ReviewAbstractFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Review
        django_get_or_create = ["moderator"]

    review_group = factory.SubFactory(ReviewGroupFactory)
    review = factory.Faker("date_time_between", start_date="+5d", end_date="+6d", tzinfo=pytz.utc)
    moderator = factory.LazyAttribute(lambda o: InternalUserFactory.create().profile)
    year = factory.Faker("pyint", min_value=1000, max_value=9999)


class ReviewFactory(ReviewAbstractFactory):
    class Meta:
        model = Review
        django_get_or_create = ["moderator", "reviewee", "reviewer"]

    from_date = factory.Faker("date_between", start_date="+2d", end_date="+3d")
    to_date = factory.Faker("date_between", start_date="+4d", end_date="+5d")
    review_deadline = factory.Faker("date_between", start_date="+5d", end_date="+6d")
    reviewee = factory.LazyAttribute(lambda o: InternalUserFactory.create().profile)
    reviewer = factory.LazyAttribute(lambda o: InternalUserFactory.create().profile)
    feedback_reviewee = factory.Faker("text")
    feedback_reviewer = factory.Faker("text")

    signed_reviewee = factory.Faker("date_time", tzinfo=pytz.utc)
    signed_reviewer = factory.Faker("date_time", tzinfo=pytz.utc)
    completely_filled_reviewee = factory.Faker("date_time", tzinfo=pytz.utc)
    completely_filled_reviewer = factory.Faker("date_time", tzinfo=pytz.utc)


class CompletedFilledReviewFactory(ReviewFactory):
    completely_filled_reviewee = factory.Faker("date_time", tzinfo=pytz.utc)
    completely_filled_reviewer = factory.Faker("date_time", tzinfo=pytz.utc)


class SignedReviewFactory(CompletedFilledReviewFactory):
    signed_reviewee = factory.Faker("date_time", tzinfo=pytz.utc)
    signed_reviewer = factory.Faker("date_time", tzinfo=pytz.utc)


class ReviewTemplateFactory(ReviewAbstractFactory):
    is_template = True


class ReviewQuestionCategoryFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = ReviewQuestionCategory

    name = factory.Faker("text", max_nb_chars=64)
    order = factory.Faker("pyint", min_value=0, max_value=9999)
    weight = factory.Faker("pydecimal", right_digits=1, min_value=0, max_value=9999)


class ReviewQuestionNoCategoryFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = ReviewQuestion

    review = factory.SubFactory(ReviewFactory)
    question = factory.Faker("text")
    order = factory.Faker("pyint", min_value=0, max_value=9999)
    weight = factory.Faker("pydecimal", right_digits=1, min_value=0, max_value=9999)


class ReviewQuestionFactory(ReviewQuestionNoCategoryFactory):
    category = factory.SubFactory(ReviewQuestionCategoryFactory)


class ReviewAnswerFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = ReviewAnswer

    question = factory.SubFactory(ReviewQuestionFactory)
    answered_by = factory.LazyAttribute(lambda o: InternalUserFactory.create().profile)
    answered_anonymized = hash(factory.SelfAttribute("answered_by"))
    answer_number = factory.Faker("pyint", min_value=0, max_value=9999)
    answer_text = factory.Faker("text")


class ReviewAnswerNoCategoryFactory(ReviewAnswerFactory):
    question = factory.SubFactory(ReviewQuestionNoCategoryFactory)


class KPIFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = KPI
        skip_postgeneration_save = True

    name = factory.Faker("pystr")
    goal = factory.Faker("pyint", min_value=0, max_value=9999)
    period = DateRange(timezone.now().date(), timezone.now().date() + timedelta(days=random.randint(4, 5)))
    handler = "wbcrm.kpi_handlers.activities.NumberOfActivityKPI"

    @factory.post_generation
    def evaluated_persons(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for person in extracted:
                self.evaluated_persons.add(person)


class DefaultPersonKPIFactory(KPIFactory):
    @factory.post_generation
    def evaluated_persons(self, create, extracted, **kwargs):
        self.evaluated_persons.add(PersonFactory())


class EvaluationFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Evaluation

    kpi = factory.SubFactory(KPIFactory)
    person = factory.SubFactory("wbcore.contrib.directory.factories.PersonFactory")
    evaluated_period = DateRange(timezone.now().date(), timezone.now().date() + timedelta(days=random.randint(4, 5)))
    evaluation_date = timezone.now().date()



from django.utils.translation import gettext_lazy as _
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.contrib.authentication.serializers import GroupRepresentationSerializer
from wbcore.contrib.directory.models import Person
from wbcore.contrib.directory.serializers import PersonRepresentationSerializer

from wbhuman_resources.models import (
    EmployeeHumanResource,
    EmployeeWeeklyOffPeriods,
    EmployeeYearBalance,
    Position,
)

from .calendars import (
    DayOffCalendarRepresentationSerializer,
    DefaultDailyPeriodRepresentationSerializer,
)


class PositionRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbhuman_resources:position-detail")

    class Meta:
        model = Position
        fields = ("id", "computed_str", "height", "level", "_detail")


class EmployeeYearBalanceRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = EmployeeYearBalance
        fields = ("id", "computed_str")


class EmployeeHumanResourceRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbhuman_resources:employee-detail")

    class Meta:
        model = EmployeeHumanResource
        fields = ("id", "computed_str", "is_active", "_detail")


class EmployeeWeeklyOffPeriodsModelSerializer(wb_serializers.ModelSerializer):
    _employee = EmployeeHumanResourceRepresentationSerializer(source="calendar")
    _period = DefaultDailyPeriodRepresentationSerializer(source="period")

    class Meta:
        model = EmployeeWeeklyOffPeriods
        fields = ("id", "employee", "period", "weekday", "_employee", "_period", "computed_str")


class PositionModelSerializer(wb_serializers.ModelSerializer):
    _parent = PositionRepresentationSerializer(source="parent")
    _manager = PersonRepresentationSerializer(source="manager", filter_params={"is_internal_profile": True})
    _groups = GroupRepresentationSerializer(source="groups", many=True)

    class Meta:
        model = Position
        read_only_fields = ["groups", "_groups"]
        fields = [
            "id",
            "name",
            "color",
            "computed_str",
            "level",
            "height",
            "parent",
            "_parent",
            "manager",
            "_manager",
            "_groups",
            "groups",
        ]


class EmployeeBalanceModelSerializer(wb_serializers.ModelSerializer):
    _profile = PersonRepresentationSerializer(source="profile")
    _position = PositionRepresentationSerializer(source="position")
    _calendar = DayOffCalendarRepresentationSerializer(source="calendar")

    available_vacation_balance_previous_year = wb_serializers.FloatField(
        label=_("Available Vacation Balance from previous year"),
        read_only=True,
        help_text=_(
            "Available Vacation Balance from previous year. Can only be used until a certain point the next year."
        ),
    )
    available_vacation_balance_current_year = wb_serializers.FloatField(
        label=_("Available Vacation Balance from current year"),
        read_only=True,
        help_text=_("Available Vacation Balance from the current year balance."),
    )

    available_vacation_balance_next_year = wb_serializers.FloatField(
        label=_("Available Vacation Balance from next year"),
        read_only=True,
        help_text=_("Available Vacation Balance from the next year balance."),
    )

    took_long_vacations = wb_serializers.BooleanField(
        label=_("Long Vacation"),
        help_text=_("True if the user took at least one long vacation in a row"),
        read_only=True,
        default=False,
    )

    extra_days_per_period = wb_serializers.FloatField(read_only=True, label=_("Extra days per period"))

    @wb_serializers.register_resource()
    def additional_resources(self, instance, request, user):
        additional_resources = dict()

        additional_resources["periods_count_per_type"] = reverse(
            "wbhuman_resources:employee-absencecount-list",
            args=[instance.id],
            request=request,
        )
        additional_resources["absencerequest"] = reverse(
            "wbhuman_resources:employee-absencerequest-list",
            args=[instance.id],
            request=request,
        )

        if (view := request.parser_context["view"]) and view.is_administrator:
            additional_resources["employeeyearbalance"] = reverse(
                "wbhuman_resources:employee-employeeyearbalance-list",
                args=[instance.id],
                request=request,
            )
        return additional_resources

    class Meta:
        model = EmployeeHumanResource
        percent_fields = ["occupancy_rate"]
        fields = [
            "id",
            "profile",
            "computed_str",
            "position",
            "_position",
            "is_active",
            "enrollment_at",
            "_profile",
            "extra_days_frequency",
            "extra_days_per_period",
            "occupancy_rate",
            "contract_type",
            "available_vacation_balance_previous_year",
            "available_vacation_balance_current_year",
            "available_vacation_balance_next_year",
            "took_long_vacations",
            "calendar",
            "_calendar",
            "_additional_resources",
        ]


class EmployeeModelSerializer(wb_serializers.ModelSerializer):
    _direct_manager = PersonRepresentationSerializer(
        source="direct_manager", filter_params={"is_internal_profile": True}
    )
    _position = PositionRepresentationSerializer(source="position")
    _calendar = DayOffCalendarRepresentationSerializer(source="calendar")
    _profile = PersonRepresentationSerializer(source="profile")

    position_manager = wb_serializers.PrimaryKeyRelatedField(read_only=True)
    _position_manager = PersonRepresentationSerializer(
        source="position_manager", filter_params={"is_internal_profile": True}
    )
    top_position_repr = wb_serializers.CharField(read_only=True)
    primary_email = wb_serializers.CharField(label=_("Primary Email"), allow_null=True, read_only=True)
    primary_address = wb_serializers.CharField(label=_("Primary Address"), allow_null=True, read_only=True)
    primary_telephone = wb_serializers.TelephoneField(label=_("Primary Telephone"), allow_null=True, read_only=True)

    @wb_serializers.register_resource()
    def extra_additional_resources(self, instance, request, user):
        res = dict()

        if instance.is_active and (view := request.parser_context["view"]):
            if view.is_administrator:
                res["deactivate"] = reverse(
                    "wbhuman_resources:employee-deactivate",
                    args=[instance.id],
                    request=request,
                )

            if (
                view.is_administrator or instance in view.employee.get_managed_employees()
            ) and instance.balances.exists():
                res["balance_and_usage"] = reverse(
                    "wbhuman_resources:employeebalance-detail",
                    args=[instance.id],
                    request=request,
                )
        return res

    class Meta:
        fields = [
            "id",
            "_profile",
            "profile",
            "position",
            "_position",
            "_position_manager",
            "position_manager",
            "top_position_repr",
            "primary_telephone",
            "primary_email",
            "primary_address",
            "direct_manager",
            "_direct_manager",
            "calendar",
            "_calendar",
            "enrollment_at",
            "occupancy_rate",
            "contract_type",
            "is_active",
            "extra_days_frequency",
            "_additional_resources",
        ]
        model = EmployeeHumanResource


class DeactivateEmployeeSerializer(wb_serializers.Serializer):
    substitute = wb_serializers.PrimaryKeyRelatedField(
        queryset=Person.objects.all(), label=_("Substitution Person"), many=False
    )
    _substitute = PersonRepresentationSerializer(source="substitute")


class EmployeeYearBalanceModelSerializer(EmployeeYearBalanceRepresentationSerializer):
    _employee = EmployeeHumanResourceRepresentationSerializer(source="employee")
    _balance = wb_serializers.FloatField(read_only=True, label="Given yearly balance (in hours)")
    _number_mandatory_days_off = wb_serializers.FloatField(read_only=True, label="Mandatory days off (in hours)")
    _total_vacation_hourly_usage = wb_serializers.FloatField(read_only=True, label="Hourly usage (in hours)")
    actual_total_vacation_hourly_balance = wb_serializers.FloatField(
        read_only=True, label="Hourly available balance (in hours)"
    )

    _balance_in_days = wb_serializers.FloatField(read_only=True, label="Given yearly balance (in days)")
    _number_mandatory_days_off_in_days = wb_serializers.FloatField(
        read_only=True, label="Mandatory days off (in days)"
    )
    _total_vacation_hourly_usage_in_days = wb_serializers.FloatField(read_only=True, label="Hourly usage (in days)")
    actual_total_vacation_hourly_balance_in_days = wb_serializers.FloatField(
        read_only=True, label="Hourly available balance (in days)"
    )

    class Meta:
        model = EmployeeYearBalance
        fields = [
            "id",
            "employee",
            "_employee",
            "extra_balance",
            "year",
            "_balance",
            "_number_mandatory_days_off",
            "_total_vacation_hourly_usage",
            "actual_total_vacation_hourly_balance",
            "_balance_in_days",
            "_number_mandatory_days_off_in_days",
            "_total_vacation_hourly_usage_in_days",
            "actual_total_vacation_hourly_balance_in_days",
            "_additional_resources",
        ]
        read_only_fields = fields



from collections import OrderedDict, defaultdict
from datetime import datetime

from dateutil import rrule
from django.utils import timezone


def get_number_of_hours_between_dates(
    d1, d2, list_employee_dayoffs, list_public_holidays=False, hours_range=None, granularity=12
):
    if hours_range is None:
        hours_range = range(0, 23)

    def convert_days_from_hours(hours, granularity, hours_per_day):
        return int(hours / granularity) * granularity / hours_per_day

    rules = rrule.rruleset()

    byweekday_list = [rrule.MO, rrule.TU, rrule.WE, rrule.TH, rrule.FR, rrule.SA, rrule.SU]

    rules.rrule(
        rrule.rrule(
            freq=rrule.HOURLY,
            byweekday=byweekday_list,
            byhour=hours_range,
            dtstart=d1,
            until=d2,
        )
    )
    if list_public_holidays:
        list_employee_dayoffs.extend(list_public_holidays)
    current_tz = timezone.get_current_timezone()
    for holiday in sorted(list_employee_dayoffs):
        s1 = datetime(holiday.year, holiday.month, holiday.day, 0, 0, 0, tzinfo=current_tz)
        s2 = datetime(holiday.year, holiday.month, holiday.day, 23, 59, 59, tzinfo=current_tz)
        rules.exrule(rrule.rrule(rrule.HOURLY, dtstart=s1, until=s2))
    dates = defaultdict(int)
    for r in list(rules):
        dates[r.date()] += 1
    final = OrderedDict()
    for k, v in dates.items():
        final[k] = convert_days_from_hours(v, granularity, len(hours_range))
    return final



import zoneinfo
from datetime import date, datetime, time
from importlib import import_module
from typing import TypeVar

import pandas as pd
from dateutil import rrule
from django.contrib.postgres.constraints import ExclusionConstraint
from django.contrib.postgres.fields import DateTimeRangeField, RangeOperators
from django.db import models
from django.db.models import CheckConstraint, F, Q, Sum
from django.db.models.signals import post_migrate
from django.dispatch import receiver
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from dynamic_preferences.registries import global_preferences_registry
from psycopg.types.range import TimestamptzRange
from timezone_field import TimeZoneField
from wbcore.contrib.agenda.models import CalendarItem
from wbcore.contrib.icons import WBIcon
from wbcore.models import WBModel

SelfDefaultDailyPeriod = TypeVar("SelfDefaultDailyPeriod", bound="DefaultDailyPeriod")


class InvalidDayOffCalendarResourceError(Exception):
    pass


class DayOffCalendar(WBModel):
    title = models.CharField(max_length=255)
    resource = models.CharField(
        max_length=255, null=True, blank=True, help_text=_("Used to fetch the days off from an API.")
    )
    timezone = TimeZoneField(default="UTC")

    def __str__(self) -> str:
        return f"{self.title}"

    def get_period_start_choices(self) -> list[str]:
        """
        Get a text choices datastructure from the possible Default Periods starts time

        Returns:
            a list of tuple containing the string representation of the periods lower time
        """
        return [str(t) for t in self.default_periods.order_by("lower_time").values_list("lower_time", flat=True)]

    def get_period_end_choices(self) -> list[str]:
        """
        Get a text choices datastructure from the possible Default Periods ends time

        Returns:
            a list of tuple containing the string representation of the periods upper time
        """
        return [str(t) for t in self.default_periods.order_by("upper_time").values_list("upper_time", flat=True)]

    def create_public_holidays(self, year: int):
        """
        Utility function that generate all holiday provided by the "workalendar" package as DayOff for the given year

        Args:
            year: Year to generate holidays from
        Raises:
            InvalidDayOffCalendarResourceError if calendar's resource is wrongly formatted.
        """
        # If no resource is specified, then we exit early
        if self.resource is None or self.resource == "":
            return

        continent, region = self.resource.split(".")
        try:
            workalendar = import_module(f"workalendar.{continent}")
            holidays = getattr(workalendar, region)()
            for holiday_date, holiday_title in holidays.holidays(year):
                DayOff.objects.get_or_create(
                    date=holiday_date, calendar=self, defaults={"count_as_holiday": True, "title": holiday_title}
                )
        except ModuleNotFoundError as e:
            raise InvalidDayOffCalendarResourceError(_("The continent you've supplied is invalid.")) from e
        except AttributeError as e:
            raise InvalidDayOffCalendarResourceError(_("The region you've supplied is invalid.")) from e

    def get_day_off_per_employee_df(self, start: date, end: date, employees) -> pd.DataFrame:
        """
        Utility function that reshape the day off for a given calendar and a list of employees among a certain date range

        Args:
            start: Start filter
            end: End filter
            employees: Queryset of employees

        Returns:
            A dataframe whose columns are [employee, period, date, type, status] with type HOLIDAY and status APPROVED
        """
        employee_df = pd.DataFrame(employees.values("id")).rename(columns={"id": "employee"})
        day_off_df = pd.DataFrame(self.days_off.filter(date__gte=start, date__lte=end).values("date"))
        periods_df = pd.DataFrame(self.default_periods.values("id")).rename(columns={"id": "period"})
        df = employee_df.merge(day_off_df, how="cross").merge(periods_df, how="cross")
        df["type"] = "Holiday"
        df["status"] = "APPROVED"
        return df

    def get_daily_hours(self) -> float:
        """
        Utility function to return the total number of hours that a typical working day spans from this calendar

        Returns:
            A float number corresponding to the total working hours
        """
        return self.default_periods.aggregate(s=Sum("total_hours"))["s"] or 0

    def get_unworked_time_range(self, start_time: time = None) -> list[tuple[int, int]]:
        """
        Utility function that return a list of unworked time range. These time range are the exclusion of the periods defined in the default period table and a typical earth time range.

        Args:
            start_time: If specified, starts the day from that time. Default to time(0,0,0)

        Returns:
            A list of time range tuple
        """
        if not start_time:
            start_time = time(0, 0, 0)
        rules = rrule.rruleset()
        rules.rrule(
            rrule.rrule(
                freq=rrule.MINUTELY,
                dtstart=datetime(1, 1, 1, 0, 0),
                until=datetime(1, 1, 1, 23, 59),
            )
        )
        for period in self.default_periods.order_by("lower_time"):
            rules.exrule(
                rrule.rrule(
                    rrule.MINUTELY,
                    dtstart=datetime(1, 1, 1, period.lower_time.hour, period.lower_time.minute, 0),
                    until=datetime(1, 1, 1, period.upper_time.hour, period.upper_time.minute, 0),
                )
            )

        non_workable_minutes = list(rules)
        pivot_index = next(x[0] for x in enumerate(non_workable_minutes) if x[1].time() >= start_time)
        non_workable_minutes = non_workable_minutes[pivot_index:] + non_workable_minutes[:pivot_index]

        start_range = None
        before_minute = None
        for minute in non_workable_minutes:
            if not start_range:
                start_range = minute
            if before_minute and minute.minute != (before_minute.minute + 1) % 60:
                yield start_range.time(), before_minute.time()
                start_range = minute
            before_minute = minute
        yield start_range.time(), before_minute.time()

    def save(self, *args, **kwargs):
        if not self.resource:
            self.resource = global_preferences_registry.manager()[
                "wbhuman_resources__calendar_default_public_holiday_package"
            ]
        if not self.timezone:
            self.timezone = zoneinfo.ZoneInfo(
                global_preferences_registry.manager()["wbhuman_resources__calendar_default_timezone"]
            )
        super().save(*args, **kwargs)

    def get_default_fullday_period(self, val_date):
        return TimestamptzRange(
            lower=self.default_periods.earliest("lower_time").get_lower_datetime(val_date),
            upper=self.default_periods.latest("lower_time").get_upper_datetime(val_date),
        )

    def normalize_period(self, period: TimestamptzRange) -> TimestamptzRange:
        """
        Given a aware range of datetime, ensure that the local time of each corresponds to a valid time choices
        Args:
            period: The period to normalize

        Returns:
            The normalize period
        """

        def _get_closest_time(periods, ts):
            return sorted(
                map(lambda x: (x[0], abs(ts - x[0].hour * 3600 + x[0].minute * 60 + x[0].second)), periods),
                key=lambda x: x[1],
            )[0][0]

        local_lower_datetime = period.lower.astimezone(self.timezone)
        local_upper_datetime = period.upper.astimezone(self.timezone)
        if str(local_lower_datetime.astimezone(self.timezone).time()) not in self.get_period_start_choices():
            closest_time = _get_closest_time(
                self.default_periods.values_list("lower_time"),
                local_lower_datetime.time().hour * 3600
                + local_lower_datetime.time().minute * 60
                + local_lower_datetime.time().second,
            )
            local_lower_datetime = datetime.combine(period.lower.date(), closest_time, tzinfo=self.timezone)
        if str(local_upper_datetime.astimezone(self.timezone).time()) not in self.get_period_end_choices():
            closest_time = _get_closest_time(
                self.default_periods.exclude(lower_time__lt=local_lower_datetime.time()).values_list("upper_time"),
                local_upper_datetime.time().hour * 3600
                + local_upper_datetime.time().minute * 60
                + local_upper_datetime.time().second,
            )
            local_upper_datetime = datetime.combine(period.upper.date(), closest_time, tzinfo=self.timezone)
        return TimestamptzRange(lower=local_lower_datetime, upper=local_upper_datetime)

    class Meta:
        verbose_name = _("Day Off Calendar")
        verbose_name_plural = _("Days Off Calendar")
        constraints = (models.UniqueConstraint(name="unique_calendar", fields=("resource", "timezone")),)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbhuman_resources:dayoffcalendar"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbhuman_resources:dayoffcalendarrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{title}}"


class DayOff(CalendarItem):
    date = models.DateField(verbose_name=_("Date"))
    count_as_holiday = models.BooleanField(
        default=True,
        verbose_name=_("Count as Holiday"),
        help_text=_("If true, there is no work but the day counts towards the employees' vacation balance"),
    )
    calendar = models.ForeignKey(
        to="wbhuman_resources.DayOffCalendar",
        related_name="days_off",
        on_delete=models.PROTECT,
        verbose_name=_("Calendar"),
    )

    class Meta:
        verbose_name = _("Day Off")
        verbose_name_plural = _("Days Off")
        constraints = [models.UniqueConstraint(fields=["date", "calendar"], name="unique_date_for_calendar")]

    def __str__(self) -> str:
        return f"{self.title} ({self.date})"

    def get_color(self) -> str:
        return "#211ae9"  # dark blue

    def get_icon(self) -> str:
        return WBIcon.DAY_OFF.icon

    def save(self, *args, **kwargs) -> TimestamptzRange:
        default_periods = self.calendar.default_periods
        self.all_day = True
        if default_periods.exists():
            self.period = self.calendar.get_default_fullday_period(self.date)
        super().save(*args, **kwargs)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbhuman_resources:dayoff"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbhuman_resources:dayoffresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{title}}"


class DefaultDailyPeriod(models.Model):
    lower_time = models.TimeField()
    upper_time = models.TimeField()

    timespan = (
        DateTimeRangeField()
    )  # "readonly" field that default to combined time with the epoch date, Used to apply database constraint
    title = models.CharField(max_length=128)
    total_hours = models.FloatField()

    calendar = models.ForeignKey(
        to="wbhuman_resources.DayOffCalendar",
        related_name="default_periods",
        on_delete=models.PROTECT,
        verbose_name=_("Calendar"),
    )

    class Meta:
        verbose_name = _("Default Daily Period")
        verbose_name_plural = _("Default Daily Periods")
        constraints = [
            CheckConstraint(condition=Q(upper_time__gt=F("lower_time")), name="check_lower_time_lt_upper_time"),
            ExclusionConstraint(
                expressions=[("timespan", RangeOperators.OVERLAPS), ("calendar", RangeOperators.EQUAL)],
                name="check_no_overlapping_default_periods_time",
            ),
        ]

    def save(self, *args, **kwargs):
        if not hasattr(self, "calendar") or not self.calendar:
            self.calendar = DayOffCalendar.objects.first()
        self.timespan = TimestamptzRange(
            lower=self.get_lower_datetime(date(1970, 1, 1)),
            upper=self.get_upper_datetime(date(1970, 1, 1)),
        )
        if not self.total_hours:
            self.total_hours = (self.timespan.upper - self.timespan.lower).total_seconds() / 3600
        super().save(*args, **kwargs)

    def get_lower_datetime(self, val_date: date) -> datetime:
        """
        Getter function to build a datetime object from this lower period time and the given date

        Args:
            val_date: The date to combine the time with

        Returns:
            A datetime
        """
        return datetime.combine(val_date, self.lower_time, tzinfo=self.calendar.timezone)

    def get_upper_datetime(self, val_date: date) -> datetime:
        """
        Getter function to build a datetime object from this uper period time and the given date

        Args:
            val_date: The date to combine the time with

        Returns:
            A datetime
        """
        return datetime.combine(val_date, self.upper_time, tzinfo=self.calendar.timezone)

    def __str__(self) -> str:
        return f"{self.title} ({self.calendar}) [{self.lower_time:%H:%M} - {self.upper_time:%H:%M}]"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbhuman_resources:defaultdailyperiodrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{title}}"


@receiver(post_migrate, sender=DayOffCalendar)
def post_migrate_day_off_calendar(sender, verbosity, interactive, stdout, using, plan, apps, **kwargs):
    """
    After migration, we check if DayOffCalendar have at least one element. OOtherwise, we create a default one
    """
    if not DayOffCalendar.objects.exists():
        calendar = DayOffCalendar.objects.create(title="Default Calendar", timezone=timezone.get_current_timezone())
        global_preferences_registry["wbhuman_resources__employee_default_calendar"] = calendar



from django.contrib import admin

from ..models import (
    DayOff,
    DayOffCalendar,
    DefaultDailyPeriod,
    EmployeeWeeklyOffPeriods,
)


@admin.register(DayOffCalendar)
class DayOffCalendarModelAdmin(admin.ModelAdmin):
    pass


@admin.register(DayOff)
class DayOffAdmin(admin.ModelAdmin):
    list_display = ["date", "title", "count_as_holiday", "calendar"]
    ordering = ["date"]


@admin.register(DefaultDailyPeriod)
class DefaultDailyPeriodAdmin(admin.ModelAdmin):
    list_display = ["lower_time", "upper_time", "title", "total_hours"]
    ordering = ["lower_time"]


class EmployeeWeeklyOffPeriodsInLine(admin.TabularInline):
    model = EmployeeWeeklyOffPeriods
    fields = [
        "period",
        "weekday",
    ]
    fk_name = "employee"
    extra = 0
    raw_id_fields = ["employee"]
    ordering = ("weekday", "period__lower_time")



from django.utils.translation import gettext as _
from wbcore.menus import ItemPermission, MenuItem

EMPLOYEEHUMANRESOURCE_MENUITEM = MenuItem(
    label=_("Balance & Usage"),
    endpoint="wbhuman_resources:employeebalance-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal,
        permissions=["wbhuman_resources.view_employeehumanresource"],
    ),
)

EMPLOYEE_MENUITEM = MenuItem(
    label=_("Employees"),
    endpoint="wbhuman_resources:employee-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal,
        permissions=["wbhuman_resources.view_employeehumanresource"],
    ),
    add=MenuItem(
        label=_("Add Employee"),
        endpoint="wbhuman_resources:employee-list",
        permission=ItemPermission(
            method=lambda request: request.user.is_internal,
            permissions=["wbhuman_resources.add_employeehumanresource"],
        ),
    ),
)

POSITION_MENUITEM = MenuItem(
    label=_("Positions"),
    endpoint="wbhuman_resources:position-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal, permissions=["wbhuman_resources.view_position"]
    ),
    add=MenuItem(
        label=_("Add Position"),
        endpoint="wbhuman_resources:position-list",
        permission=ItemPermission(
            method=lambda request: request.user.is_internal, permissions=["wbhuman_resources.add_position"]
        ),
    ),
)



import factory
from django.contrib.auth.models import Group
from django.db.models.signals import post_save
from faker import Faker

from wbhuman_resources.models import (
    BalanceHourlyAllowance,
    EmployeeHumanResource,
    EmployeeWeeklyOffPeriods,
    EmployeeYearBalance,
    Position,
)

fake = Faker()


class PositionFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Position
        skip_postgeneration_save = True

    name = factory.Faker("text", max_nb_chars=64)
    manager = None

    @factory.post_generation
    def post_gen(self, create, extracted, **kwargs):
        group = Group.objects.get_or_create(name="Test Group")[0]
        self.groups.add(group)


@factory.django.mute_signals(post_save)
class EmployeeHumanResourceFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = EmployeeHumanResource
        django_get_or_create = ["profile"]
        skip_postgeneration_save = True

    is_active = True
    profile = factory.SubFactory("wbcore.contrib.authentication.factories.AuthenticatedPersonFactory")
    direct_manager = factory.SubFactory("wbcore.contrib.directory.factories.PersonFactory")
    calendar = factory.SubFactory("wbhuman_resources.factories.DayOffCalendarFactory")
    position = factory.SubFactory(PositionFactory)
    enrollment_at = factory.Faker("past_date")
    extra_days_frequency = EmployeeHumanResource.ExtraDaysBalanceFrequency.YEARLY

    contract_type = EmployeeHumanResource.ContractType.INTERNAL
    occupancy_rate = factory.Faker("pyfloat", min_value=0, max_value=1)


class EmployeeYearBalanceFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = EmployeeYearBalance
        django_get_or_create = ["employee", "year"]
        skip_postgeneration_save = True

    employee = factory.SubFactory("wbhuman_resources.factories.EmployeeHumanResourceFactory")
    year = factory.Faker("pyint", min_value=1970, max_value=2022)
    extra_balance = factory.Faker("pyint", min_value=0, max_value=20)

    @factory.post_generation
    def post_gen(self, create, extracted, **kwargs):
        if create:
            BalanceHourlyAllowanceFactory.create(balance=self)


class BalanceHourlyAllowanceFactory(factory.django.DjangoModelFactory):
    balance = factory.SubFactory("wbhuman_resources.factories.EmployeeYearBalanceFactory")
    period_index = 1
    hourly_allowance = factory.Faker("pyint", min_value=160, max_value=200)

    class Meta:
        model = BalanceHourlyAllowance
        django_get_or_create = ["balance", "period_index"]


class EmployeeWeeklyOffPeriodsFactory(factory.django.DjangoModelFactory):
    employee = factory.SubFactory("wbhuman_resources.factories.EmployeeHumanResourceFactory")
    period = factory.LazyAttribute(lambda x: x.employee.calendar.default_periods.order_by("?")[0])
    weekday = factory.Faker("pyint", min_value=0, max_value=7)

    class Meta:
        model = EmployeeWeeklyOffPeriods
        django_get_or_create = ["employee", "period", "weekday"]



from django.utils.translation import gettext as _
from wbcore.menus import ItemPermission, MenuItem

REVIEWGROUP_MENUITEM = MenuItem(
    label=_("Groups"),
    endpoint="wbhuman_resources:reviewgroup-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal, permissions=["wbhuman_resources.view_reviewgroup"]
    ),
    add=MenuItem(
        label=_("Add Group"),
        endpoint="wbhuman_resources:reviewgroup-list",
        permission=ItemPermission(
            method=lambda request: request.user.is_internal, permissions=["wbhuman_resources.add_reviewgroup"]
        ),
    ),
)


REVIEW_MENUITEM = MenuItem(
    label=_("Employee Reviews"),
    endpoint="wbhuman_resources:review-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal, permissions=["wbhuman_resources.view_review"]
    ),
    add=MenuItem(
        label=_("Add Review"),
        endpoint="wbhuman_resources:review-list",
        permission=ItemPermission(
            method=lambda request: request.user.is_internal, permissions=["wbhuman_resources.add_review"]
        ),
    ),
)

REVIEWTEMPLATE_MENUITEM = MenuItem(
    label=_("Templates"),
    endpoint="wbhuman_resources:reviewtemplate-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal, permissions=["wbhuman_resources.add_review"]
    ),
    add=MenuItem(
        label=_("Add Template"),
        endpoint="wbhuman_resources:reviewtemplate-list",
        permission=ItemPermission(
            method=lambda request: request.user.is_internal, permissions=["wbhuman_resources.add_review"]
        ),
    ),
)

REVIEWQUESTIONCATEFORY_MENUITEM = MenuItem(
    label=_("Categories"),
    endpoint="wbhuman_resources:reviewquestioncategory-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal,
        permissions=["wbhuman_resources.view_reviewquestioncategory"],
    ),
    add=MenuItem(
        label=_("Add Question Category"),
        endpoint="wbhuman_resources:reviewquestioncategory-list",
        permission=ItemPermission(
            method=lambda request: request.user.is_internal,
            permissions=["wbhuman_resources.add_reviewquestioncategory"],
        ),
    ),
)

REVIEWQUESTION_MENUITEM = MenuItem(
    label=_("Questions"),
    endpoint="wbhuman_resources:reviewquestion-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal, permissions=["wbhuman_resources.view_reviewquestion"]
    ),
    add=MenuItem(
        label=_("Add Question"),
        endpoint="wbhuman_resources:reviewquestion-list",
        permission=ItemPermission(
            method=lambda request: request.user.is_internal, permissions=["wbhuman_resources.add_reviewquestion"]
        ),
    ),
)


REVIEWANDWER_MENUITEM = MenuItem(
    label=_("Answers"),
    endpoint="wbhuman_resources:reviewanswer-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal, permissions=["wbhuman_resources.view_reviewanswer"]
    ),
    add=MenuItem(
        label=_("Add Review"),
        endpoint="wbhuman_resources:reviewanswer-list",
        permission=ItemPermission(
            method=lambda request: request.user.is_internal, permissions=["wbhuman_resources.add_reviewanswer"]
        ),
    ),
)
